#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['qpsolvers', 'qpsolvers.solvers']

package_data = \
{'': ['*']}

install_requires = \
['quadprog >=0.1.8', 'scipy >=1.2.0']

extras_require = \
{'all_pypi_solvers': ['cvxopt >=1.2.6',
                      'cvxpy >=1.1.11',
                      'ecos >=2.0.7.post1',
                      'osqp >=0.6.2',
                      'scs >=2.1.2,<3.0'],
 'doc': ['sphinx']}

setup(name='qpsolvers',
      version='1.7.2',
      description='Quadratic programming solvers in Python with a unified API',
      author=None,
      author_email='Stéphane Caron <stephane.caron@normalesup.org>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
